#!/usr/bin/perl
# use -w only while testing
# 
# mwiki (aka paxifixi)
#
# MAIL BASED CONTENT MANAGEMENT SYSTEM
#
# authors: Max-Gerd Retzlaff <m.retzlaff@gmx.net>,
#          Reyk Floeter <reyk@synack.de>,
#          Sebastian Horzela <jules@killall5.de>
# date:    $Date: 2002/03/07 01:34:51 $
# version: $Revision: 1.7 $
#
# This application is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version. 
#
# This application is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Library General Public License for more details.
#
# You should have received a copy of the GNU General Public
# License along with this library; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# 02111 USA.  

# changing to dir in which mwiki.pl is stored
BEGIN {
    (my $dir = $0) =~ s/^(.*)\/[^\/]*?$/$1/;
    chdir( $dir );
}

###################################################################
# default settings, do() config file, modules, usage information
#
# modules
use structuredText;
use HelpfulFunctions; 
use MailParser;
use generateHTML;
use GnuPGverify;
use Mail::Mailer;

#default# $local_config_dir = "/etc/mwiki";
$local_config_dir = "../etc/mwiki";
$local_config_file = $local_config_dir."/mwiki_local.conf";

$dist_config_file = $local_config_dir."/mwiki_dist.conf";
$dist_sections_conf_file = $local_config_dir."/sections_dist.conf";
$dist_main_body_file = $local_config_dir."/main_body_dist.html";
$dist_skin_dir = $local_config_dir."/skins";
# The config file is a piece of perl code now.
# You have (at least) to define:
#    - $top_dir
#    - $www_dir
#    - $mwiki_addr
# You should also define:
#    - $main_title
#    - $sub_title
#    - $index_title
#    - $main_body_file

$use_pointer = "yes";
$dir_perm_public   = 0755;
$file_perm_public  = 0644;
$file_perm_private = 0600;

### default values that should be overwritten in your config file (see above)
# now testing whether $top_dir is defined, so the next two lines are uncommented
# $top_dir = "nil";  
# $www_dir = "nil";  
$mwiki_addr = "mwiki\@localhost";

$latest_news_number = 5;

$main_title = "Contents";
$sub_title = "Contents of";
$index_title = "Index";

%sections = ();
# $sections_conf_file = '';

$main_body = "\n<h2>Welcome to mwiki</h2>\n<em>MAIL BASED CONTENT MANAGEMENT SYSTEM</em>\n";
$main_body_file = ''; # -> not present
###

# execute config file 
# this has to happen AFTER the assignment of the default values
do $local_config_file;

if ( (!defined $top_dir) or
     (!defined $www_dir) or
     (!defined $config_dir) or
     (!defined $config_file) ) {
    mdie(1, "ERROR: Invalid configuration! Check \"$local_config_file\".\n");
}

if (defined $config_file and -f $config_file) {
    do $config_file;
} else {
    do $dist_config_file;
}

if (defined $sections_conf_file and -f $sections_conf_file) {
    read_sections_conf( $sections_conf_file, \%sections );
} else {
    read_sections_conf( $dist_sections_conf_file, \%sections );
}

generateHTML->set_www_dir($www_dir); # has to be invoked after "do $local_config_file"
if(defined $www_skin_dir) {
    generateHTML->set_www_skin_dir($www_skin_dir); # has to be invoked after "do $local_config_file"
}
file_as_string( \$main_body, $main_body_file );

$latest_news_file = "$top_dir/latest_news.inc";
$index_inc_file   = "$top_dir/index.inc";
$pointer_inc_file = "$top_dir/pointer.inc";

# prepare usage information
$usage = <<"END_OF_USAGE";

usage:\tmwiki.pl -[m|M|i|p|r|R|S]

\t-m  process message from stdin
\t-M  process message from stdin and generate pointers and indexes
\t-i  generate indexes
\t-p  make pointers
\t-r  re-generate html content
\t-R  generate indexes and pointers and re-generate html content
\t-S  create subdirectories and (empty) include files in \$top_dir,
\t      then generate indexes and pointers and re-generate html content
\t      (IOW: create a skeleton page / this is non-destructive)

\t note: You would normally use only the big-letter options.

status:
\t\$top_dir:\t$top_dir
\t\$www_dir:\t$www_dir
\t\$local_config_file:\t$local_config_file
\t\$config_file:\t$config_file

END_OF_USAGE



###################################################################
# parse program parameters
if ($#ARGV != 0) {
    mdie(1, $usage);
}

if ($ARGV[0] eq "-m" or $ARGV[0] eq "-M") {
    my $stdin = '';
    {
	local $/ = undef;
	$stdin = <STDIN>;
    }
    parse_mail($stdin, \%sections);
    if ($ARGV[0] eq "-M") {
	make_pointers();
	generate_indexes(\%sections);
    }
} elsif ($ARGV[0] eq "-i") {
    generate_indexes(\%sections);
} elsif ($ARGV[0] eq "-r") {
    generate_content(\%sections);
} elsif ($ARGV[0] eq "-p") {
    make_pointers();
} elsif ($ARGV[0] eq "-R") {
    generate_indexes(\%sections);
    make_pointers();
    generate_content(\%sections);
} elsif ($ARGV[0] eq "-S") {
    create_skeleton(\%sections);

    if (defined $config_file and -f $config_file) {
	do $config_file;
    }
    read_sections_conf( $sections_conf_file, \%sections );
    file_as_string( \$main_body, $main_body_file );

    generate_indexes(\%sections);
    make_pointers();
    generate_content(\%sections);
} else {
    mdie(1, $usage);
}


###################################################################
# read sections.conf
sub read_sections_conf {
    my ($filename, $sections_rh) = @_;

    if ($filename ne '' and -f $filename) {

	open(SOURCE, "< $filename") or 
	    mdie(1, "ERROR: Can't open file \"$filename\" for reading.\n");
	my @sections_conf = <SOURCE>;
	close(SOURCE);
	
	my $section = undef;
	foreach my $line (@sections_conf) {
	    if ( $line =~ /^name:\s*([^\n]*?)\s*$/i ) {
		$section = $1;
		$sections_rh->{$section}->{AllowSimpleMail} = 'NO';
	    } elsif ( defined $section ) {
		if ( $line =~  /^aliases:\s*([^\n]*)$/i ) {
		    @aliases = split( /(?:\s*,\s*|\s+)/ ,  $1 );
		    # quite hackish: /(?:\s+|\s*,\s*)/ would not run!
		    foreach my $alias (@aliases) {
			push(@{$sections_rh->{$section}->{aliases}}, uc($alias) );
		    }
		} elsif ( $line =~ /^[^\#]*(ADM|ADD)\s+SimpleMail/i ) {
		    $sections_rh->{$section}->{AllowSimpleMail} = $1;
		} elsif ( $line =~ /^[^\#]*ADM\s+(0x[0-9A-Fa-f]{8})/i ) {
		    push(@{$sections_rh->{$section}->{adm_IDs}}, $1 );
		} elsif ( $line =~ /^[^\#]*ADD\s+(0x[0-9A-Fa-f]{8})/i ) {
		    push(@{$sections_rh->{$section}->{add_IDs}}, $1 );
		} # else { ignore lines that are not understood }
	    }
	}
    }
    return 0;
}



###################################################################
# generate subdirectories and (empty) include files in $top_dir
# create_sekelton() is non-destructive. This meas that you can
# add categories (see %sections) to an existing mwiki installation.
sub create_skeleton {
    my %sections = %{$_[0]};

    foreach my $dir (keys %sections) {
	mkdir( "$top_dir/$dir" );
	chmod( $dir_perm_public, "$top_dir/$dir" );
    }

    string_to_file(">".$latest_news_file, '', $file_perm_private);
    string_to_file(">".$index_inc_file, '', $file_perm_private);
    string_to_file(">".$index_inc_file, '', $file_perm_private);
    string_to_file(">".$pointer_inc_file, '', $file_perm_private);

    mkdir( $config_dir );
    chmod( $dir_perm_public, $config_dir );
    
    foreach my $file ('config', 'sections_conf', 'main_body') {
	my $text = '';	
	file_as_string( \$text, ${'dist_'.$file.'_file'} );
        string_to_file( ${$file.'_file'}, $text, $file_perm_private )
    if (! -f ${$file.'_file'} );
    } 

    if(defined $skin_dir and ! -d $skin_dir) {
        if( -d $dist_skin_dir ) {
            system("/bin/cp", "-a", $dist_skin_dir, $skin_dir);
            system("/bin/chmod", "-R", "u=rwX,og=rX", $skin_dir);
        }
    }
	
    return 0;
}



###################################################################
# generates directory indexes
sub generate_indexes {
    my %sections = %{$_[0]};

    my $latest_news = '';
    file_as_string( \$latest_news, $latest_news_file );

    my $main .= "\t<a href='".$www_dir."/index.html'>".$main_title."</a><p>\n";
    foreach my $cwd ( sort(keys(%sections)) ) {
	$main .= "\t<li><a href=\"$www_dir/$cwd/index.html\">$cwd</a></li>\n";
    }
 
    if ($use_pointer eq "yes") {
	$main .= "\t<p><li><a href=\"".$www_dir."/pointer.html\">pointer</a></li>";
    }

    # write MAIN MENU
    string_to_file("$index_inc_file", $main, $file_perm_private);

    # write $top_dir."/index.html"
    string_to_file($top_dir."/index.html",
		   generateHTML->whole_page($main_title, '', $index_title, $main,
					    $main_title, $main_body,
					    $latest_news),
		   $file_perm_public);

    # generate indexes for subdirectories
    foreach my $cwd ( sort(keys(%sections)) ) {
	opendir(MYDIR, $top_dir."/".$cwd);
    
	# write index for subdirectory
	my $htmltext = '';

	while($filename = readdir(MYDIR)) {
	    if ($filename =~ /\.txt$/) {
		my $rawmail = '';
		file_as_string( \$rawmail, $top_dir."/".$cwd."/".$filename );
		
		my $mail = MailParser->autodetectingnew($rawmail, \%sections);

		my $from = $mail->get_sender();
		my (undef, undef, undef, $title) = $mail->parse_subject();
		my $date = $mail->get_header_field( 'Date' );

		(my $filename_html = $filename ) =~ s/(.*)\.txt/$1\.html/;

		$htmltext .= "\t<li><a href='".$filename_html."'>".$title."</a>&nbsp\n";
		$htmltext .= "<br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;\n";
		$htmltext .= "(by <font size='-1'><b>".$from."</b> on ".$date.")&nbsp; ";
		$htmltext .= "<a href='".$filename."'>[src]</a></font></li>\n";
	    }
	}

	string_to_file($top_dir."/".$cwd."/index.html", 
		       generateHTML->whole_page($cwd, '', $index_title, $main,
						"$sub_title $cwd",
						"<ul>$htmltext</ul>", $latest_news),
		       $file_perm_public);
	closedir(MYDIR);
    }
    return 0;
}



###################################################################
# create $topdir."/pointer.html"
sub make_pointers
{
    my $index = '';
    file_as_string( \$index, $index_inc_file );

    my $pointers = '';
    file_as_string( \$pointers, $pointer_inc_file);

    $pointers = join("\n", unique_array( split("\n", $pointers) ) );

    my $htmltext = structuredText2HTML("\n".$pointers);

    string_to_file($top_dir."/pointer.html",
		   generateHTML->whole_page("Pointers", '', $index_title, $index,
					    "Pointers", $htmltext, 
					    scalar localtime() ),
		   $file_perm_public );

    return 0;
}



###################################################################
# re-creates content files
sub generate_content
{
    my %sections = %{$_[0]};
    my $filename;

    ## delete pointer file
    # unlink $pointer_inc_file;
    #  this would not work as intended
    #  because links of refused mails cannot be regained
    #  -> new, third option for parse_mail()
    
    foreach my $cwd ( sort(keys(%sections)) ) {
	opendir(MYDIR, $top_dir."/".$cwd);
 
	while($filename = readdir(MYDIR)) {
	    if ($filename =~ /\.txt/) {
		my $input = '';
		file_as_string( \$input, $top_dir."/".$cwd."/".$filename );
		parse_mail($input, \%sections, "re-generation (do not search for links)");
	    }
	}
	
	closedir(MYDIR);
    }
    return 0;
}



###################################################################
# processes incomming messages
sub parse_mail
{
    my $rawmail = $_[0];
    my %sections = %{$_[1]};
    my $regeneration = defined($_[2]) ? '0' : '1';

    if ($regeneration) {
	my @pointer = $rawmail =~ m/(\"[^\"]*?\"\s\((?:http|https|ftp|news|telnet):.*?\))/gis;
	
	foreach my $link (@pointer) { # correct messed links, strip =\n
	    $link =~ s/\=\n//g;
	}

        # alternative:
	# my @pointer = ();
	# while ( $rawmail =~ m/(\"[^\"]*?\"\s\((http|https|ftp|news|telnet|mailto):.*?\))/gis ) {
	# push( @pointer, $1);
	# # print STDERR "->$1<-\n";
        # }
	# print join("\n", @pointer)."\n";

	string_to_file(">".$pointer_inc_file, 
		       join('', map("- $_\n", @pointer) ), # join() is needed!
		       $file_perm_private);

	make_pointers(); # <- is a workaround, cf. TODO file
    }

    my $mail = MailParser->autodetectingnew($rawmail, \%sections);

    my ($command, $alias, $mid, $title) = $mail->parse_subject();
    # print "-$command-$alias-$mid-$title-\n";

    # send digest
    if ( $command eq 'SENDDIGEST' ) {
	send_digest( $alias, $mid, $title, \%sections );
	return 0; # go out of parse_mail()
    }

    # process digest
    if ( $command eq 'DIGEST' ) {
	process_digest( $rawmail, \%sections );
	return 0; # go out of parse_mail()
    }

    $mid = $mail->get_mid() if ( $mid eq '' );
    # use a random filename for missing Message-IDs...
    $mid = time()."-".rand(1).'@no.where' if ( $mid eq '');

    # $mail->replace_subject_and_mid( $command, $alias, $mid, $title );
    $mail->replace_header_field( "Subject", "ADD: $alias: $title" );
    (my $mid_tmp = $mid) =~ s/-at-/\@/;
    $mail->replace_header_field( "Message-ID", "<$mid_tmp>" );

    my $section = undef;
    foreach my $valid_section (keys %sections) {
	foreach my $valid_alias (@{$sections{$valid_section}->{aliases}} ) {
	    if ($alias eq $valid_alias) {
		$section = $valid_section;
	    }
	}
    }
    
    if (!defined $section) {
	mdie(0,"ERROR: unsupported mail alias (Don't support messages of alias $alias.)\n");
    }

    # convert structured Text to HTML
    my ($mail_content, $sign_ID) = $mail->get_content( \%sections );
    my $htmlmail = structuredText2HTML( $mail_content );
    mdie(1,"ERROR: Couldn't transform structured Text to html.\n") if ($htmlmail eq ""); # should not happen... ;-)

#   if ($sign_ID ne 'SimpleMail') {
    $mail->replace_header_field( "X-mwiki-sign-ID", "$sign_ID" );
#   }

    $filename = $top_dir."/".$section."/mwiki-".$mid;

    my $old_from = '';
    my $old_sign_ID = '';
    my $old_incoming_date = '';
    my $old_date = '';
    # check whether the sender is allowed to update/upgrade
    if ( $command eq 'UPDATE' or
	 $command eq 'UPGRADE' or
	 $command eq 'REPLACE' or
	 $command eq 'DELETE' or
	 $command eq 'REMOVE') {
	
	if ($sign_ID eq 'SimpleMail') {
	    if ( $sections{$section}->{AllowSimpleMail} ne 'ADM' ) {
		mdie(0, "ERROR: AllowSimpleMail is not set to ADM in section \"$section\".\n");
	    }
	} else {
	    
	    my $raw_old_mail = '';
	    file_as_string( \$raw_old_mail, "$filename.txt" );
	    if ( $raw_old_mail ne '' ) {
		my $old_mail = MailParser->autodetectingnew($raw_old_mail, \%sections);
		$old_sign_ID = $old_mail->get_header_field( 'X-mwiki-sign-ID' );
		$old_from = $old_mail->get_header_field( 'From' );
		$old_date = $old_mail->get_header_field( 'Date' );
		$old_incoming_date = $old_mail->get_header_field( 'X-mwiki-incoming-date' );
		
		my $sign_ID_is_adm = 0;
		foreach my $valid_ID (@{$sections{$section}->{adm_IDs}}) {
		    if ( $valid_ID eq $sign_ID ) {
			$sign_ID_is_adm = 1;
		    }
		}
		
		if ( $sign_ID ne $old_sign_ID and $sign_ID_is_adm == 0 ) {
		    mdie(0, "ERROR: ".$sign_ID." wanted to update/upgrade/remove an message signed by ID ".$old_sign_ID." and is no adm_ID in section \"$section\".\n");
		}
	    }
	}
    }

    if ( $command eq 'UPDATE' ) {
	if ( $old_from ne '' ) {
	    $mail->replace_header_field( 'From', $old_from );
	}
	if ( $old_sign_ID ne '') {
	    $mail->replace_header_field( "X-mwiki-sign-ID", "$old_sign_ID" );
	}
	if ( $old_date ne '' ) {
	    $mail->replace_header_field( 'Date', $old_date );
	}
	if ( $old_incoming_date ne '' ) {
	    $mail->replace_header_field( "X-mwiki-incoming-date", $old_incoming_date, "STAMP" );
	} else {
	    $mail->replace_header_field( "X-mwiki-incoming-date", time(), "STAMP" );
	}
    }
    

    my $incoming_date = $mail->get_header_field( 'X-mwiki-incoming-date' );
    if ($incoming_date eq '') {
	$incoming_date = time();
    }
    my $date = $mail->get_header_field( 'Date' );
    my $from = $mail->get_sender();

    my $index = "";
    file_as_string( \$index, $index_inc_file);

	# write last updated information - new: only when a new mail is processed
    if ($regeneration and $command ne 'UPDATE') {
	my $latest_news = "";
	file_as_string( \$latest_news, $latest_news_file );
	my @latest_news = split("\n", $latest_news);

	if ( $command eq "DELETE" or $command eq "REMOVE" ) {
	    my @new_latest_news = ();
	    foreach my $news (@latest_news) {
		if ( $news !~ /$section\/mwiki-$mid.html/ ) {
		    push( @new_latest_news, $news);
		}
	    }
	    @latest_news = @new_latest_news;
	} else {
	    unshift( @latest_news, "<em><a href=\"$www_dir/".$section."/index.html\">".$section."</a>-&gt;<a href=\"$www_dir/".$section."/mwiki-$mid.html\">".$title."</a></em><br>" );
#	    unshift( @latest_news, "<em><a href=\"$www_dir/".$section."/index.html\">".$section."</a>-&gt;<a href=\"$www_dir/".$section."/mwiki-$mid.html\">".$title."</a></em>, ".localtime()."<br>" );
	}

	string_to_file($latest_news_file,
		       join("\n", @latest_news[0..($latest_news_number-1)] ),
		       $file_perm_private );
    }

    my $special = "<p/ ><font size='-1'>";
    $special.= "[<a href='mwiki-".$mid.".txt'>view mail source</a>]&nbsp;\n";
    $special.= "[<a href='mailto:".$mwiki_addr."?subject=UPDATE: ".$alias.": [".$mid."] ".$title."'>update page</a>]&nbsp;\n";
    $special.= "[<a href='mailto:".$mwiki_addr."?subject=UPGRADE: ".$alias.": [".$mid."] ".$title."'>upgrade page</a>]&nbsp;\n";
    $special.= "[<a href='mailto:".$mwiki_addr."?subject=REMOVE: ".$alias.": [".$mid."] ".$title."'>remove page</a>]&nbsp;\n";
    $special.= "$date</font>";


    if ($command eq "DELETE" or
	$command eq "REMOVE") {
	unlink "$filename.html", "$filename.txt";
    } else {
	# write HTML'ized output
	string_to_file($filename.".html",
		       generateHTML->whole_page($title, '', $index_title, $index,
			    "<a href=\"$www_dir/".$section."/index.html\">".
			    $section." -&gt;</a> ".$title.
			    " &nbsp;<i><small>(by $from)</small></i>",
			    $htmlmail. $special, scalar localtime($incoming_date) ),
		       $file_perm_public );
	# write mail source
	# string_to_file($filename.".txt", $rawmail, $file_perm_public );
	string_to_file($filename.".txt", $$mail, $file_perm_public );
    }

    return 0;
}



###################################################################
# generates directory indexes
sub send_digest {
    my ($alias, $mid, $title) = @_[0..2];
    my %sections = %{$_[3]};

    my $digest = '';

    # append mails to digest
    foreach my $cwd ( sort(keys(%sections)) ) {    

	my $append_this_dir = undef;
	foreach my $valid_alias (@{$sections{$cwd}->{aliases}}, 'MWIKIALL' ) {
	    if ($alias eq $valid_alias) {
		$append_this_dir = 1;
	    }
	}

	if (defined $append_this_dir) {
	    opendir(MYDIR, $top_dir."/".$cwd);
    
	    while($filename = readdir(MYDIR)) {
		if ($filename =~ /\.txt$/) {
		    my $rawmail = '';
		    file_as_string( \$rawmail, $top_dir."/".$cwd."/".$filename );
		    
		    $digest .= $rawmail."\n\n";
		}
	    }
	    closedir(MYDIR);
	}
    }

    # mask ^From\ 's
    $digest =~ s/(^|\n)(\>*From\ )/$1>$2/g;

    # send digest as mail to $mid
    $mid =~ s/-at-/\@/;
    eval {
	my $mailer = Mail::Mailer->new();
	$mailer->open( { "From"    => $mwiki_addr,
			 "To"      => $mid,
			 "Subject" => "DIGEST: $alias: your title was: $title"
#			 "Content-Type" => "text/plain;"
			 } ) or mdie(1, "ERROR: Couldn't open \$mailer: $!.\n");
	print $mailer $digest;
	$mailer->close() or mdie(1, "ERROR: Couldn't close \$mailer: $!.\n");
    };
    if ($@) {
	mdie(1, "ERROR: Couldn't send digest to $mid. Eval error: $@.\n");
    }

    return 0;
}



###################################################################
# processes incomming messages
sub process_digest {
    my $rawmail = $_[0];
    my %sections = %{$_[1]};

    my (undef, $content) =  split/\n\n/, $rawmail, 2;

    # unmask ^From\ 's
    $content =~ s/(^|\n)(\>)(\>*From\ )/$1$3/g;

    $content =~ s/(\n\nFrom\ )/$1From\ /g;
    my @mails = split("\n\n+From\ ", $content);

    foreach my $mail (@mails) {
	parse_mail($mail, \%sections);
    }	    

    return 0;
}



###################################################################
exit 0;
